/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <libetpan/cinthash.h>

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"
#include "folder.h"
#include "lxml.h"
#include "processing.h"
#include "manager.h"
#include "mailbox.h"
#include "message.h"
#include "vmessage.h"

#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
#include "xep-python.h"
#endif

typedef struct XePFolderAppendMessagesData_s XePFolderAppendMessagesData;
struct XePFolderAppendMessagesData_s
{
  XePFolderAppendMessagesCb user_cb;
  void *user_data;
  carray *messages;
  carray *folders;
  uint32_t count;
  uint32_t index;
  uint32_t folder_count;
  uint32_t failures;
};
typedef struct XePFolderAppendData_s XePFolderAppendData;
struct XePFolderAppendData_s
{
  XePFolderAppendMessagesData *data;
  XePMessage *message;
  uint32_t count;
};

typedef struct XePFolderListCopyCbData_s XePFolderListCopyCbData;
struct XePFolderListCopyCbData_s
{
  XePFolderListCopyCb user_cb;
  void *user_data;
  unsigned int count;
};

struct XePFolderProcessing_s
{
  XePFolder *folder;
  carray *messages;
  clist *processing_rules;
  clist *matches;
};

static XePFolder *
Folder_new(XePFolderType type);

static void
Folder_free(XePFolder *folder);

static void
Folder_list_copy_free_copy_cb(void *acb_data, void *afolder);

static void
Folder_list_copy_finalize(XePFolderListCopyCbData *cb_data);

static int
Folder_ref_list_copy(clist *source, clist *target);

static int
Folder_ep_storage_create(XePFolder *folder);

static int
Folder_ep_folder_create(XePFolder *folder);

static void 
Folder_update_msg_session(XePFolder *folder);

static void
Folder_append_next_message_async(XePFolderAppendMessagesData *data);

static void
Folder_append_next_message_fetched_cb(void *adata, void *amessage);

static void
Folder_append_next_message_done_cb(void *aappend_data, void *aappend_message);

static void
Folder_append_file_cb(void *anull, void *amessage);

static uint32_t
Folder_tree_get_total_count(struct mailmessage_tree *msg_tree);

#define XEP_FOLDER_VIRTUAL_INITIAL_SIZE 10

#define XEP_FOLDER_RECONNECT_MAX_TRIES 5

#define XEP_FOLDER_HASH_SIZE 128
static cinthash_t *folder_uid_hash_table;
static chash *folder_ref_hash_table;
static clist *global_folder_refs;


/* Function definitions */

XePFolder *
a_Folder_new(XePFolderType type)
{
  XePFolder *folder;
  static long uid = 0;

  A_LOGGING_TRACE;

  folder = Folder_new(type);
  if (folder)
    {
      /* TODO: remove the periodicity of folder uid, because in
       * the very long term, in theory, two folders can have the
       * same uid
       */
      if (uid < 0) /* Use only the positive space */
	uid = 0;
      folder->uid          = ++uid;

      if (folder_uid_hash_table == NULL)
	folder_uid_hash_table = cinthash_new(XEP_FOLDER_HASH_SIZE);

      cinthash_add(folder_uid_hash_table, folder->uid, folder);
    }
  return folder;
}

static XePFolder *
Folder_new(XePFolderType type)
{
  XePFolder *folder;

  folder = x_alloc(1, XePFolder);

  if (folder == NULL)
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
  else
    {

      folder->id           = NULL;
      folder->type         = type; 
      folder->uid          = 0;
      folder->thread.state = XEP_MANAGER_THREAD_STATE_NONE;
      folder->mailbox      = NULL;
      folder->path         = NULL;
      folder->ep_storage   = NULL;
      folder->ep_folder    = NULL;
      folder->ep_tree      = NULL;
      folder->ep_list      = NULL;
      folder->total        = 0;
      folder->unseen       = 0;
      folder->recent       = 0;

      switch (type)
	{
	case XEP_FOLDER_TYPE_MAILBOX:
	  /* We can also define mailbox processing rules,
	   * so we initialize a mailbox-type folder as a real-type one:
	   */
	case XEP_FOLDER_TYPE_REAL:
	  folder->rule_refs = a_Processing_ref_list_new();
	  folder->folder_refs = NULL;
	  folder->condition_refs = NULL;
	  folder->msg_attrs = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);
	  if (folder->msg_attrs == NULL)
	    goto err;

	  break;
	case XEP_FOLDER_TYPE_VIRTUAL:
	  folder->rule_refs = NULL;
	  folder->folder_refs = a_Folder_ref_list_new();
	  folder->condition_refs = a_Processing_ref_list_new();
	  folder->vmessages = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);
	  folder->msg_attrs = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);
	  folder->ep_list = x_alloc(1, struct mailmessage_list);
	  if (folder->ep_list)
	    {
	      folder->ep_list->msg_tab = 
		carray_new(XEP_FOLDER_VIRTUAL_INITIAL_SIZE);
	      if (folder->ep_list->msg_tab == NULL)
		goto err;
	    }
	  else
	    goto err;
	  break;
	default:
	  break;
	}
    }
  return folder;
 err:
  a_Folder_free(folder);
  return NULL;
}

XePFolder *
a_Folder_new_copy(XePFolder *folder)
{
  XePFolder *folder_copy;
  int r = 0;

  folder_copy = Folder_new(folder->type);
  if (folder_copy == NULL)
    {
      return NULL;
    }
  
  folder_copy->uid  = folder->uid;
  folder_copy->id   = X_STRDUP(folder->id);
  folder_copy->path = X_STRDUP(folder->path);
  
  switch (folder_copy->type)
    {
    case XEP_FOLDER_TYPE_REAL:
      r = a_Processing_ref_list_copy(folder->rule_refs,
				     folder_copy->rule_refs);

      break;
    case XEP_FOLDER_TYPE_VIRTUAL:

      r = Folder_ref_list_copy(folder->folder_refs, folder_copy->folder_refs);

      if (r == 0)
	r = a_Processing_ref_list_copy(folder->condition_refs,
				       folder_copy->condition_refs);

      break;
    default:
      break;
    }
  
  if (r == -1)
    {
      a_Folder_free_copy(folder_copy);
      folder_copy = NULL;
    }

  return folder_copy;
}

void
a_Folder_free(XePFolder *folder)
{
  if (folder == NULL)
    return;

  cinthash_remove(folder_uid_hash_table, folder->uid);
  Folder_free(folder);
}

void
a_Folder_free_copy(XePFolder *folder)
{
  if (folder)
    Folder_free(folder);
}

static void
Folder_free(XePFolder *folder)
{
  A_LOGGING_TRACE;

  x_free(folder->id);

  x_free(folder->path);

  switch (folder->type)
    {
    case XEP_FOLDER_TYPE_REAL:
      a_Processing_ref_list_free(folder->rule_refs);
      /* TODO: free all attributes */
      chash_free(folder->msg_attrs);
      break;
    case XEP_FOLDER_TYPE_VIRTUAL:
      a_Processing_ref_list_free(folder->condition_refs);
      a_Folder_ref_list_free(folder->folder_refs);
      /* TODO: Remove all vmessages */
      chash_free(folder->vmessages);
      /* TODO: free all attributes */
      chash_free(folder->msg_attrs);
      break;
    default:
      break;
    }

  x_free(folder);
}


void
a_Folder_copy(XePFolder *source, XePFolder *target)
{
  int r = 0;

  x_free(target->id);
  target->id = X_STRDUP(source->id);

  x_free(target->path);
  target->path = X_STRDUP(source->path);

  /* Type and UID cannot be changed by the user, so they are the
   * same 
   */

  /* The mailbox is certainly the same, too */


  switch (source->type)
    {
    case XEP_FOLDER_TYPE_MAILBOX:
    case XEP_FOLDER_TYPE_REAL:
      a_Processing_ref_list_free(target->rule_refs);
      target->rule_refs = clist_new();
      r = a_Processing_ref_list_copy(source->rule_refs,
				     target->rule_refs);

      break;
    case XEP_FOLDER_TYPE_VIRTUAL:
      a_Folder_ref_list_free(target->folder_refs);
      target->folder_refs = clist_new();
      r = Folder_ref_list_copy(source->folder_refs, target->folder_refs);

      if (r == 0)
	{
	  a_Processing_ref_list_free(target->condition_refs);
	  target->condition_refs = clist_new();
	  r = a_Processing_ref_list_copy(source->condition_refs,
					 target->condition_refs);
	}

      break;
    default:
      break;
    }
#if 0
  /* TODO: handle errors */
     return r;
#endif
}

XePFolder *
a_Folder_get_from_uid(long uid)
{
  return (XePFolder *) cinthash_find(folder_uid_hash_table, uid);
}

const char *
a_Folder_get_id(XePFolder *folder)
{
  return folder->id;
}

void 
a_Folder_set_path(XePFolder *folder, const char *path)
{
  x_free(folder->path);
  folder->path = path ? strdup(path) : NULL;
  /* Invalidate id */
  if (folder->id)
    {
      free(folder->id);
      folder->id = NULL;
    }
}

void
a_Folder_set_path_p(XePFolder *folder, char *path)
{
  x_free(folder->path);
  folder->path = path;
  if (folder->id)
    {
      free(folder->id);
      folder->id = NULL;
    }
}

const char *
a_Folder_get_path(XePFolder *folder)
{
  return folder->path;
}

void
a_Folder_set_mailbox(XePFolder *folder, XePMailbox *mailbox)
{
  folder->mailbox = mailbox;
  /* Invalidate id */
  if (folder->id)
    {
      free(folder->id);
      folder->id = NULL;
    }
}

XePMailbox *
a_Folder_get_mailbox(XePFolder *folder)
{
  return folder->mailbox;
}

void
a_Folder_update_id(XePFolder *folder)
{
  const char *mailbox_id;
  size_t mailbox_id_len0;
  size_t folder_path_len0;

  /* vfolders do not have an associated mailbox, but their "path" is their 
   * name
   */
  if (folder->type == XEP_FOLDER_TYPE_VIRTUAL)
    mailbox_id = "";
  else
    mailbox_id = a_Mailbox_get_id(folder->mailbox);

  if (folder->path == NULL)
    {
      folder->id = strdup(mailbox_id);
    }
  else
    {
      mailbox_id_len0 = strlen(mailbox_id) + 1;
      folder_path_len0 = strlen(folder->path) + 1;
      folder->id = x_alloc(mailbox_id_len0 + folder_path_len0, char);
      memcpy(folder->id, mailbox_id, mailbox_id_len0);
      folder->id[mailbox_id_len0 - 1] = ':';
      memcpy(folder->id + mailbox_id_len0, folder->path, folder_path_len0);
    }
}

const char *
a_Folder_get_short_name(XePFolder *folder)
{
  const char *p;

  if (folder->type == XEP_FOLDER_TYPE_VIRTUAL)
    return folder->path;
  if (folder->type == XEP_FOLDER_TYPE_MAILBOX)
    return folder->mailbox->id;

  switch (folder->mailbox->type)
    {
    case XEP_MAILBOX_IMAP:
      /* FIXME: Folder should have a hierarchy separator member */
      if (((p = strrchr(folder->path, ':')) == NULL) && 
	  ((p = strrchr(folder->path, '.')) == NULL) &&
	  ((p = strrchr(folder->path, '/')) == NULL))
	return folder->path;
      else 
	return ++p;
      break;
    case XEP_MAILBOX_MH:
      /* FIXME: Folder should have a hierarchy separator member */
      if ((p = strrchr(folder->path, '/')) == NULL)
	return folder->path;
      else
	return ++p;
      break;
    default:
      return folder->path;
    }
}

clist *
a_Folder_list_new(void)
{
  clist *folders;
 
  folders = clist_new();

  if (folders == NULL)
      A_LOGGING_SYSTEM_ERROR("clist_new failed");

  return folders;
}

clist *
a_Folder_list_new_copy(clist *folders)
{
  clist *folders_copy;
  clistiter *iter;
  XePFolder *folder;
  XePFolder *folder_copy;
  int r = 0;

  folders_copy = clist_new();
  if (folders_copy == NULL)
    {
      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }

  for (iter = clist_begin(folders); iter && r == 0; iter = iter->next)
    {
      folder = (XePFolder *) iter->data;
      folder_copy = a_Folder_new_copy(folder);
      if (folder_copy)
	{
	  if (clist_append(folders_copy, folder_copy))
	    {
	      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
	      r = -1;
	    }
	}
      else
	{
	  r = -1;
	}
    }
  if (r == -1)
    {
      a_Folder_list_free_copy(folders_copy);
      folders_copy = NULL;
    }
  return folders_copy;
}

void
a_Folder_list_free(clist *folders)
{
  clistiter *iter;

  if (!folders)
    return;

  A_LOGGING_TRACE;

  for (iter = clist_begin(folders); iter; iter = iter->next)
      a_Folder_free((XePFolder *) iter->data);

  clist_free(folders);    
}

void
a_Folder_list_free_copy(clist *folders)
{
  clistiter *iter;

  if (!folders)
    return;

  A_LOGGING_TRACE;

  for (iter = clist_begin(folders); iter; iter = iter->next)
      a_Folder_free_copy((XePFolder *) iter->data);

  clist_free(folders);    
}

void
a_Folder_list_copy(clist *source, clist *target,
		   XePMailbox *mailbox,
		   XePFolderListCopyCb user_cb, void *user_data)
{
  clistiter *iter;
  clistiter *next;
  int r = 0;
  XePFolder *folder_copy;
  XePFolder *folder;
  XePFolderListCopyCbData *cb_data;

  cb_data = x_alloc(1, XePFolderListCopyCbData);
  if (cb_data == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      /* FIXME: add a return code to check when malloc fails */
      user_cb(user_data);
      return;
    }

  cb_data->user_cb = user_cb;
  cb_data->user_data = user_data;
  cb_data->count = 1;
  
  iter = clist_begin(target);
  while (iter)
    iter = clist_delete(target, iter);

  iter = clist_begin(source);
  while (iter && r == 0)
    {
      folder_copy = (XePFolder *) iter->data;
      next = iter->next;

      if (folder_copy->uid < 0)
	{
	  /* Folder was marked for delete */
	  folder = a_Folder_get_from_uid(-folder_copy->uid);
	  if (folder)
	    {
	      a_Folder_unregister(folder);
	      cb_data->count++;
	      a_Folder_disconnect_async(folder,
					Folder_list_copy_free_copy_cb,
					cb_data);
	    }
	  else
	    {
	      A_LOGGING_SYSTEM_ERROR("attempt to delete a non-existent "
				     "folder %s", folder_copy->id);
	    }
	}
      else
	{
	  folder = a_Folder_get_from_uid(folder_copy->uid);
	  if (folder == NULL)
	    {
	      /* A new unregistered folder */
	      /* Steal the copy and register it. */
	      folder = folder_copy;
	      a_Folder_set_mailbox(folder, mailbox);
	      a_Folder_update_id(folder);
	      a_Folder_register(folder);
	      /* Be kind enough, and replace the stolen folder
	       * with a copy */
	      folder_copy = a_Folder_new_copy(folder);
	      iter->data = folder_copy;
	    }
	  else if (folder == folder_copy)
	    {
	      /* a new registered folder */
	      folder = a_Folder_new_copy(folder_copy);
	      iter->data = folder;
	      folder = folder_copy;
	      a_Folder_set_mailbox(folder, mailbox);
	      a_Folder_update_id(folder);
	    }
	  else
	    {
	      /* a modified folder */
	      a_Folder_copy(folder_copy, folder);
	      a_Folder_set_mailbox(folder, mailbox);
	      /* FIXME: when ID changes, the ref hash table has to
	       * be updated
	       */
	      a_Folder_update_id(folder);	      
	    }
	  clist_append(target, folder);
	}
      iter = next;
    }
  Folder_list_copy_finalize(cb_data);
}

static void
Folder_list_copy_free_copy_cb(void *acb_data, void *afolder)
{
  XePFolderListCopyCbData *cb_data = (XePFolderListCopyCbData *) acb_data;
  XePFolder *folder = (XePFolder *) afolder;

  a_Folder_free(folder);
  Folder_list_copy_finalize(cb_data);
}
  
static void
Folder_list_copy_finalize(XePFolderListCopyCbData *cb_data)
{
  XePFolderListCopyCb user_cb;
  void *user_data;

  cb_data->count--;
  if (cb_data->count)
    return;

  user_cb = cb_data->user_cb;
  user_data = cb_data->user_data;

  free(cb_data);

  user_cb(user_data);
}

#if 0
static void
Folder_list_dump(XePFoldersList *folderslist)
{
  clistiter *iter;
  for (iter = clist_begin(folderslist); iter; iter = iter->next)
    {
      XePFolder *folder = iter->data;
      printf("Folder type %d id = %s, Mailbox = %s, Path = %s\n",
	     folder->type, folder->id, folder->mailbox->id, folder->path);
    }
}
#endif

XePFolderRef *
a_Folder_ref_new(void)
{
  XePFolderRef *folder_ref;
  
  folder_ref = x_alloc(1, XePFolderRef);

  if (folder_ref == NULL)
    A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
  else
    {
      folder_ref->ref = NULL;
      folder_ref->folder = NULL;
      folder_ref->enabled = 1;
    }
  if (global_folder_refs == NULL)
    global_folder_refs = clist_new();

  clist_append(global_folder_refs, folder_ref);

  return folder_ref;      
}

XePFolderRef *
a_Folder_ref_new_copy(XePFolderRef *folder_ref)
{
  XePFolderRef *folder_ref_copy;
  
  folder_ref_copy = a_Folder_ref_new();

  if (folder_ref_copy)
    {
      folder_ref_copy->ref = X_STRDUP(folder_ref->ref);
      folder_ref_copy->enabled = folder_ref->enabled;
    }

  return folder_ref_copy;
}

void
a_Folder_ref_free(XePFolderRef *folder_ref)
{
  clistiter *iter;

  if (folder_ref == NULL)
    return;

  for (iter = clist_begin(global_folder_refs); 
       iter && iter->data != folder_ref;
       iter = iter->next);

  if (iter)
    {
      clist_delete(global_folder_refs, iter);
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("freeing an unregistered XePFolderRef %s",
			     folder_ref->ref);
    }

  a_Folder_ref_free_content(folder_ref);


  x_free(folder_ref);
}

void
a_Folder_ref_free_content(XePFolderRef *folder_ref)
{
  if (folder_ref == NULL)
    return;

  x_free(folder_ref->ref);

  folder_ref->ref = NULL;
  folder_ref->folder = NULL;
}

static int
Folder_ref_list_copy(clist *source, clist *target)
{
  XePFolderRef *folder_ref;
  XePFolderRef *folder_ref_copy;

  clistiter *iter;
  int r = 0;

  for (iter = clist_begin(source); iter && r == 0; iter = iter->next)
    {
      folder_ref = (XePFolderRef *) iter->data;
      folder_ref_copy = a_Folder_ref_new_copy(folder_ref);
      if (folder_ref_copy)
	{
	  if (clist_append(target, folder_ref_copy))
	    {
	      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
	      a_Folder_ref_free(folder_ref_copy);
	      r = -1;
	    }
	}
      else
	{
	  r = -1;
	}
    }
  return r;
}

void
a_Folder_ref_relink_all(void)
{
  clistiter *iter;
  XePFolderRef *folder_ref;
  for (iter = clist_begin(global_folder_refs);
       iter;
       iter = iter->next)
    {
      folder_ref = (XePFolderRef *) iter->data;
      a_Folder_ref_link_to_folder(folder_ref);
    }
}

int
a_Folder_ref_link_to_folder(XePFolderRef *folder_ref)
{
  int r;

  if (folder_ref_hash_table == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("folder_ref_hash_table is NULL");
      return -1;
    }

  if (folder_ref == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("folder_ref is NULL");
      return -1;
    }
  
  if (folder_ref->ref == NULL)
    {
      A_LOGGING_DEBUG("Folder reference is empty.");
      return -1;
    }

  folder_ref->folder = a_Folder_get_from_ref(folder_ref->ref);

  r = (folder_ref->folder == NULL);
		  
  return r;
}

clist *
a_Folder_ref_list_new(void)
{
  return clist_new();
}

void
a_Folder_ref_list_free(clist *folder_refs)
{
  clistiter *iter;

  if (folder_refs == NULL)
    return;

  for (iter = clist_begin(folder_refs); iter; iter = iter->next)
    a_Folder_ref_free((XePFolderRef *)iter->data);

  clist_free(folder_refs);
}

chash *
a_Folder_ref_hash_new(void)
{
  folder_ref_hash_table = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);

  if (folder_ref_hash_table == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("chash_new failed");
    }

  return folder_ref_hash_table;
}

void
a_Folder_ref_hash_free(void)
{
  chash_free(folder_ref_hash_table);
}

chash *
a_Folder_ref_hash_get(void)
{
  return folder_ref_hash_table;
}

int
a_Folder_register(XePFolder *folder)
{
  int r = -1;
  chashdatum key;
  chashdatum data;

  A_LOGGING_DEBUG("registering folder %s", folder->id);

  key.data = folder->id;
  key.len = strlen(folder->id);

  if (chash_get(folder_ref_hash_table, &key, &data) == 0)
    {
      A_LOGGING_ERROR("A folder with id='%s' is already here. "
		      "New folder ignored.", folder->id);
    }
  else 
    {
      data.data = (char *) folder;
      data.len  = sizeof(folder);
      if (chash_set(folder_ref_hash_table, &key, &data, NULL))
	{
	  A_LOGGING_SYSTEM_ERROR("chash_set failed on folder '%s'. "
				 "Folder not registred.", folder->id);
	}
      else 
	{
	  r = 0;
	}
    }
  return r;
}

int
a_Folder_unregister(XePFolder *folder)
{
  int r = 0;
  chashdatum key;
  chashdatum data;

  A_LOGGING_DEBUG("unregistering folder %s", folder->id);

  key.data = folder->id;
  key.len = strlen(folder->id);

  if (chash_get(folder_ref_hash_table, &key, &data) == 0)
    {
      chash_delete(folder_ref_hash_table, &key, &data);
    }
  else 
    {
      A_LOGGING_SYSTEM_ERROR("no registered folder id='%s' to unregister.",
			     folder->id);
      r = -1;
    }
  return r;
}

int
a_Folder_register_list(clist *folders)
{
  int r = 0;
  clistiter *iter;
  XePFolder *folder;

  A_LOGGING_TRACE;

  for (iter = clist_begin(folders); iter && !r; iter = iter->next)
    {
      folder = (XePFolder *) iter->data;
      r = a_Folder_register(folder);
    }
  return r;
}

int
a_Folder_unregister_list(clist *folders)
{
  int r = 0;
  clistiter *iter;
  XePFolder *folder;

  for (iter = clist_begin(folders); iter && !r; iter = iter->next)
    {
      folder = (XePFolder *) iter->data;
      r = a_Folder_unregister(folder);
    }
  return r;
}

XePFolder *
a_Folder_get_from_ref(const char *ref)
{
  chashdatum key;
  chashdatum data;
  XePFolder *folder = NULL;

  A_LOGGING_TRACE;

  key.data = (char *) ref;
  key.len = strlen(ref);

  if (chash_get(folder_ref_hash_table, &key, &data) == 0)
    {
      folder = (XePFolder *) data.data;
    }

  return folder;
}


XePManagerThreadState
a_Folder_get_thread_state(XePFolder *folder)
{
  return a_Manager_thread_get_state_safe(&folder->thread);
}

int
a_Folder_is_locked(XePFolder *folder)
{
  return a_Manager_thread_data_is_locked(&folder->thread);
}

/*********************************************************************/
/*                                                                   */
/* Folder Operations                                                 */
/* These are in general called from within a thread                  */
/*                                                                   */
/*********************************************************************/

int
a_Folder_init(XePFolder *folder)
{
  int r;

  A_LOGGING_TRACE;

  if (folder->ep_storage)
    {
      A_LOGGING_SYSTEM_ERROR("folder->ep_storage already initialized");
      return -1;
    }

  if (folder->ep_folder)
    {
      A_LOGGING_SYSTEM_ERROR("folder->ep_folder already initialized");
      return -1;
    }

  r = Folder_ep_storage_create(folder);

  if (r == MAIL_NO_ERROR)
    r = Folder_ep_folder_create(folder);

  return r;
}

static int
Folder_ep_storage_create(XePFolder *folder)
{ 
  int r;
  struct mailstorage *ep_storage;
  XePMailbox *mailbox = NULL;

  A_LOGGING_TRACE;

  mailbox = folder->mailbox;

  ep_storage = mailstorage_new(NULL);
  if (ep_storage == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("mail_storage_new failed");
      return  -1;
    }
  switch (mailbox->type)
    {
    case XEP_MAILBOX_IMAP:
      r = imap_mailstorage_init(ep_storage, 
				mailbox->server->host,
				mailbox->server->port,
				NULL,
				CONNECTION_TYPE_PLAIN,
				IMAP_AUTH_TYPE_PLAIN,
				mailbox->server->login,
				mailbox->server->password,
				(mailbox->cache != NULL), mailbox->cache);
      if (r != MAIL_NO_ERROR)
	{
	  A_LOGGING_ERROR("Error while initializing IMAP storage");
	  mailstorage_free(ep_storage);
	  return -1;
	}
      break;
    case XEP_MAILBOX_MBOX:
      /* TODO: Path should be the concatenation of mailbox->toplevel and
       * folder->path
       */
      r = mbox_mailstorage_init(ep_storage, 
				folder->path,
				(mailbox->cache != NULL),
				mailbox->cache,
				mailbox->flags);
      if (r != MAIL_NO_ERROR)
	{
	  A_LOGGING_ERROR("Error while initializing MBOX storage");
	  mailstorage_free(ep_storage);
	  return -1;
	}
      break;
    case XEP_MAILBOX_MH:
      r = mh_mailstorage_init(ep_storage, 
			      folder->path,
			      (mailbox->cache != NULL),
			      mailbox->cache,
			      mailbox->flags);
      if (r != MAIL_NO_ERROR)
	{
	  A_LOGGING_ERROR("Error while initializing MH storage");
	  mailstorage_free(ep_storage);
	  return -1;
	}
      break;
    case XEP_MAILBOX_NNTP:
      r = nntp_mailstorage_init(ep_storage, 
				mailbox->server->host,
				mailbox->server->port,
				NULL,
				CONNECTION_TYPE_PLAIN,
				NNTP_AUTH_TYPE_PLAIN,
				mailbox->server->login,
				mailbox->server->password,
				(mailbox->cache != NULL), 
				mailbox->cache, 
				mailbox->flags);
      if (r != MAIL_NO_ERROR)
	{
	  A_LOGGING_ERROR("Error while initializing NNTP storage");
	  mailstorage_free(ep_storage);
	  return -1;
	}
      break;
    case XEP_MAILBOX_POP3:
      r = pop3_mailstorage_init(ep_storage, 
				 mailbox->server->host,
				 mailbox->server->port,
				 NULL,
				 CONNECTION_TYPE_PLAIN,
				 IMAP_AUTH_TYPE_PLAIN,
				 mailbox->server->login,
				 mailbox->server->password,
				 (mailbox->cache != NULL),
				 mailbox->cache, 
				 mailbox->flags);
      if (r != MAIL_NO_ERROR)
	{
	  A_LOGGING_ERROR("Error while initializing POP3 storage");
	  mailstorage_free(ep_storage);
	  return -1;
	}
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("attempt to connect to an unintialized "
			     "mailbox type for folder '%s'", folder->id);
    }

  folder->ep_storage = ep_storage;

  return 0;
}

static int
Folder_ep_folder_create(XePFolder *folder)
{
  struct mailfolder *ep_folder;

  ep_folder = mailfolder_new(folder->ep_storage, folder->path, NULL);

  if (ep_folder == NULL)
    {
      A_LOGGING_ERROR("Error while initializing folder");
      mailstorage_free(folder->ep_storage);
      folder->ep_storage = NULL;
      return -1;
    }
  folder->ep_folder = ep_folder;

  return 0;
}

int
a_Folder_connect(XePFolder *folder)
{
  int r;
  int reconnect;
  int connect_tries = 0;
  mailsession *old_session;

  A_LOGGING_TRACE;

  if (folder->ep_folder == NULL)
    {
      r = a_Folder_init(folder);
      if (r != MAIL_NO_ERROR)
	return r;
    }

  A_LOGGING_DEBUG("connecting to folder");
 
  old_session = folder->ep_folder->fld_session;

  reconnect = 1;
  do
    {
      r = mailfolder_connect(folder->ep_folder);
      if (r == MAIL_NO_ERROR)
	{
	  r = mailsession_noop(folder->ep_folder->fld_session);
	  if (r == MAIL_ERROR_NOT_IMPLEMENTED)
	    r = MAIL_NO_ERROR;
	  if (r == MAIL_NO_ERROR)
	    {
	      reconnect = 0;
	      mailsession_status_folder(folder->ep_folder->fld_session,
					folder->ep_folder->fld_pathname,
					&folder->total,
					&folder->recent,
					&folder->unseen);
	    }
	}

      if (reconnect)
	{
	  connect_tries++;
	  if (folder->ep_folder->fld_shared_session)
	    {
	      mailfolder_disconnect(folder->ep_folder);
	      mailstorage_disconnect(folder->ep_folder->fld_storage);
	    }
	  else
	    {
	      mailfolder_disconnect(folder->ep_folder);
	    }
	}
    } while (reconnect && connect_tries < XEP_FOLDER_RECONNECT_MAX_TRIES);

  if (reconnect)
    {	
      A_LOGGING_ERROR("error connecting folder '%s': %s", folder->id,
		      maildriver_strerror(r));
    }
  else if (old_session != folder->ep_folder->fld_session)
    {
      Folder_update_msg_session(folder);
    }

  return r;
}

static void 
Folder_update_msg_session(XePFolder *folder)
{
  mailmessage *msg;
  carray *msg_array;
  uint32_t i;

  A_LOGGING_TRACE;
  if (!folder || !folder->ep_list)
    return;

  msg_array = folder->ep_list->msg_tab;
  if (msg_array == NULL)
    return;

  for (i = 0; i < msg_array->len; i++)
    {
      msg = (mailmessage *) carray_get(msg_array, i);

      if (msg)
	{
	  msg->msg_session = folder->ep_folder->fld_session;
	  if (msg->msg_driver == imap_cached_message_driver)
	    {
	      struct imap_cached_session_state_data *ancestor_session;
	      mailmessage *ancestor_msg;
	      
	      ancestor_session = folder->ep_folder->fld_session->sess_data;
	      ancestor_msg = msg->msg_data;
	      ancestor_msg->msg_session = ancestor_session->imap_ancestor;
	    }
	}
    }
}

int
a_Folder_disconnect(XePFolder *folder)
{

  if (folder->ep_tree)
    {
      mailmessage_tree_free_recursive(folder->ep_tree);
      folder->ep_tree = NULL;
    }
  if (folder->ep_list)
    {
      mailmessage_list_free(folder->ep_list);
      folder->ep_list = NULL;
    }

  if (folder->ep_folder)
    {
      mailfolder_free(folder->ep_folder);
      folder->ep_folder = NULL;
    }

  if (folder->ep_storage)
    {
      /* This also does disconnection */
      mailstorage_free(folder->ep_storage);
      folder->ep_storage = NULL;
    }
  
  return 0;
}

int
a_Folder_noop(XePFolder *folder)
{
  int r;

  A_LOGGING_TRACE;

  if (folder->ep_folder == NULL)
    {
      r = a_Folder_connect(folder);
      /* a_Folder_connect() already does a NOOP */
      return r; 
    }

  A_LOGGING_DEBUG("session NOOP");

  r = mailsession_noop(folder->ep_folder->fld_session);
  if (r == MAIL_ERROR_NOT_IMPLEMENTED)
    {
      r = MAIL_NO_ERROR;
    }

  if (r == MAIL_NO_ERROR || (r = a_Folder_connect(folder)) == MAIL_NO_ERROR)
    {
      mailsession_status_folder(folder->ep_folder->fld_session,
				folder->ep_folder->fld_pathname,
				&folder->total,
				&folder->recent,
				&folder->unseen);
    }
  return r;
}

int
a_Folder_expunge(XePFolder *folder)
{
  int r;

  A_LOGGING_TRACE;

  r = a_Folder_connect(folder);
  if (r != MAIL_NO_ERROR)
    return r;

  return mailsession_expunge_folder(folder->ep_folder->fld_session);
}

int
a_Folder_sync(XePFolder *folder)
{
  mailsession *session;

  /* Just testing...*/
  /* TODO: check memory leaks... */
  struct mailmessage_list * env_list;
  struct mailmessage_tree * env_tree;

  int r;

  A_LOGGING_TRACE;

  r = a_Folder_connect(folder);
  if (r != MAIL_NO_ERROR)
    return r;
  
  session = folder->ep_folder->fld_session;

  A_LOGGING_TIME("get_messages_list Start");
  r = mailsession_get_messages_list(session, &env_list);
  A_LOGGING_TIME("get_messages_list Stop");
  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("Folder '%s': error while getting message list: %s ",
		      folder->id,
		      maildriver_strerror(r));
      a_Folder_disconnect(folder);
      return r;
    }

  printf("\ngetting env list\n");
  A_LOGGING_TIME("get_envelopes_list Start");
  r = mailsession_get_envelopes_list(session, env_list);
  A_LOGGING_TIME("get_envelopes_list Stop");
  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("folder '%s': error envelopes list: %s",
		      folder->id,
		      maildriver_strerror(r));
      mailmessage_list_free(env_list);
      a_Folder_disconnect(folder);
      return r;
    }

  printf("\nbuilding threads\n");
  /* FIXME: Do not hardcode the encoding (here UTF8) */
  r = mail_build_thread(MAIL_THREAD_REFERENCES_NO_SUBJECT,
			"UTF-8", 
			env_list, 
			&env_tree,
			mailthread_tree_timecomp);

  printf("\nbuilding thread done\n");

  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("can't build tree");
      mailmessage_list_free(env_list);
      a_Folder_disconnect(folder);
      return r;
    }

  if (folder->ep_tree)
    mailmessage_tree_free_recursive(folder->ep_tree);
  if (folder->ep_list)
    mailmessage_list_free(folder->ep_list);

  folder->ep_tree = env_tree;
  folder->ep_list = env_list;

  return r;
}

int
a_Folder_check(XePFolder *folder)
{
  int r;
  
  A_LOGGING_TRACE;

  if (folder->ep_folder == NULL)
    {
      r = a_Folder_connect(folder);
      if (r != MAIL_NO_ERROR)
	return r;
    }

  A_LOGGING_TIME("Folder_check Start");
  r = mailsession_check_folder(folder->ep_folder->fld_session);
  if (r == MAIL_ERROR_NOT_IMPLEMENTED)
    {
      r = MAIL_NO_ERROR;
    }
  A_LOGGING_TIME("Folder_check Stop");

  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("Error while checking folder: %s",
		      maildriver_strerror(r));
    }
  return r;
}

int
a_Folder_append(XePFolder *folder, const char *data, size_t size)
{
  int r;

  /* NOOP helps synchronising the server before an APPEND in
   * case the IMAP MAILBOX was modified by another client
   */
  r = a_Folder_noop(folder);
  if (r != MAIL_NO_ERROR)
    return r;

  r = mailsession_append_message(folder->ep_folder->fld_session, data, size);
  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("Error while appending a message: %s", 
		      maildriver_strerror(r));
    }
  return r;
}

int
a_Folder_mailbox_connect(XePFolder *folder)
{
  int r;
  int reconnect;
  int connect_tries = 0;

  A_LOGGING_TRACE;
  

  if (folder->ep_storage)
    return MAIL_NO_ERROR;

  r = Folder_ep_storage_create(folder);
  if (r != MAIL_NO_ERROR)
    return r;

  reconnect = 1;

  do
    { 
      r = mailstorage_connect(folder->ep_storage);
      if (r != MAIL_NO_ERROR)
	{
	  connect_tries++;
	  mailstorage_disconnect(folder->ep_storage);
	}
      else
	{
	  reconnect = 0;
	}
    } while (reconnect && connect_tries < XEP_FOLDER_RECONNECT_MAX_TRIES);

  if (reconnect)
    {
      A_LOGGING_ERROR("error connecting mailbox '%s': %s", folder->id,
		      maildriver_strerror(r));
    }

  return r;
}

int
a_Folder_mailbox_disconnect(XePFolder *folder)
{

  if (folder->ep_storage == NULL)
    return 0;

  mailstorage_disconnect(folder->ep_storage);
  mailstorage_free(folder->ep_storage);
  folder->ep_storage = NULL;

  return 0;
}

int
a_Folder_mailbox_list(XePFolder *folder, struct mail_list **mb_list)
{
  int r;
  char *reference = "";

  r = a_Folder_mailbox_connect(folder);
  if (r != MAIL_NO_ERROR)
    return r;

  r = mailsession_list_folders(folder->ep_storage->sto_session, 
			       reference, 
			       mb_list);
  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("Error while retrieving list of folders: %s",
		      maildriver_strerror(r));
    }
  return r;
}

int
a_Folder_mailbox_lsub(XePFolder *folder, struct mail_list **mb_list)
{
  int r;
  char *reference = "";

  r = a_Folder_mailbox_connect(folder);
  if (r != MAIL_NO_ERROR)
    return r;

  r = mailsession_lsub_folders(folder->ep_storage->sto_session,
			       reference,
			       mb_list);
  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("Error while retrieving list of folders: %s",
		      maildriver_strerror(r));
    }
  return r;
}

/* Hi-Level functions which submit job request to threads */

int
a_Folder_array_append_messages_async(carray *folders, carray *messages,
				     XePFolderAppendMessagesCb user_cb,
				     void *user_data)
{
  uint32_t count = carray_count(messages);
  XePFolderAppendMessagesData *data;

  /* This algorithm is taylored for a list of messages located on
   * one folder. If the messages come from different folders,
   * and if it is not possible to call it with a list of messages
   * from same folder, then it would be better if the messages
   * are sorted round-robin according to their location:
   * messages[0] from folder0
   * ...
   * messages[n] from foldern
   * messages[n+1] from folder0
   * Nevertheless, the destination folder will always be 
   * the last bottleneck
   */

  A_LOGGING_DEBUG("Copying %d messages to %d folders", 
		  count, carray_count(folders));

  if (count == 0)
    {
      if (user_cb)
	user_cb(folders, messages, 0, user_data);
      return 0;
    }

  data  = x_alloc(1, XePFolderAppendMessagesData);
  if (data == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      if (user_cb)
	user_cb(folders, messages, 0, user_data);
      return -1;
    }

  data->folders = folders;
  data->messages = messages;
  data->count = 1;
  data->index = 0;
  data->user_cb = user_cb;
  data->user_data = user_data;
  data->failures = 0;

#if 0
  /* TODO: we now have an array of destination folders... */
  if (0 /* FIXME: disabled */ && folder->mailbox->type == XEP_MAILBOX_IMAP)
    {
      /* Sort from the list messages that are in the same mailbox
       * and use special IMAP command for them
       */
      return 0;
    }
#endif
  A_LOGGING_TIME("Append messages Start");
  Folder_append_next_message_async(data);
  return 0;
}

static void
Folder_append_next_message_async(XePFolderAppendMessagesData *data)
{
  carray *messages = data->messages;
  XePManagerMsg *msg;
  XePMessage *message;

  A_LOGGING_TRACE;

  /* amessage == NULL => start fetching */
  /* amessage != NULL => append finished, fetch next */

  message = (XePMessage *) carray_get(messages, data->index);

  data->index++;

  /* Fetch RFC2822 message from source folder. 
   * The callback is responsible to append the fetched message
   * to the destination folder
   */
  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_FETCH);
  a_Manager_msg_set_args(msg, message, message);
  a_Manager_msg_set_source_target(msg, NULL, &message->folder->thread);
  a_Manager_msg_add_callback(msg, Folder_append_next_message_fetched_cb, data);
  a_Manager_msg_submit(msg);
}

static void
Folder_append_next_message_fetched_cb(void *adata, void *amessage)
{
  XePFolderAppendMessagesData *data = (XePFolderAppendMessagesData *) adata;
  XePMessage *message = (XePMessage *) amessage;
  XePFolderAppendData *append_data;

  A_LOGGING_TRACE;

  append_data = x_alloc(1, XePFolderAppendData);

  if (append_data && message->data)
    {
      XePManagerMsg *msg;
      uint32_t i;
      uint32_t folder_count = carray_count(data->folders);
      XePMessage *append_message;

      append_data->data = data;
      append_data->count = 0;
      append_data->message = message;

      A_LOGGING_DEBUG("Appending msg %d", message->msg->msg_index);

      for (i = 0; i < folder_count; i++)
	{
	  XePFolderRef *folder_ref = 
	    (XePFolderRef *) carray_get(data->folders, i);
	  if (folder_ref->folder == NULL)
	    {
	      /* TODO: do not ignore silently */
	      continue;
	    }
	  /* Change message folder to the destination folder */
	  append_message = a_Message_new();
	  if (append_message == NULL)
	    {
	      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
	      continue;
	    }
	  append_message->msg_data = message->data;
	  append_message->msg_size = message->len;
	  append_message->folder = folder_ref->folder;

	  data->count++;
	  append_data->count++;

	  msg = a_Manager_msg_new();
	  a_Manager_msg_set_type(msg, XEP_MANAGER_APPEND);
	  a_Manager_msg_set_args(msg, append_message, append_message);
	  a_Manager_msg_set_source_target(msg, NULL, 
					  &append_message->folder->thread);
	  a_Manager_msg_add_callback(msg, Folder_append_next_message_done_cb, 
				     append_data);
	  a_Manager_msg_submit(msg);
	}
      if (append_data->count == 0)
	{
	  x_free(append_data);
	  append_data = NULL;
	}
    }
  if (append_data == NULL || message->data == NULL)
    {
      data->failures++;
      A_LOGGING_ERROR(_("Message %d from folder %s could not be appended."),
		      message->msg->msg_index,
		      a_Folder_get_id(message->folder));

      /* TODO: call a user callback on failure for real time notification */

      Folder_append_next_message_done_cb(NULL, message);
    }

  if (data->index < carray_count(data->messages))
    Folder_append_next_message_async(data);
  else
    data->count--;
}

static void
Folder_append_next_message_done_cb(void *aappend_data, void *aappend_message)
{
  XePFolderAppendData *append_data  = (XePFolderAppendData *) aappend_data;
  XePFolderAppendMessagesData *data = append_data->data;
  XePMessage *append_message = (XePMessage *) aappend_message;
  XePMessage *message;
  XePFolderAppendMessagesCb user_cb;
  void *user_data;
  carray *messages;
  carray *folders;
  uint32_t failures;
  
  A_LOGGING_TRACE;

  if (aappend_data)
    {
      append_data->count--;
      data->count--;
      A_LOGGING_DEBUG("append_data->count = %d data->count = %d", 
		      append_data->count, data->count);

      if (append_data->count)
	return;

      append_message->data = NULL;
      a_Message_unref(append_message);
      message = append_data->message;
      A_LOGGING_DEBUG("Msg %d appended", message->msg->msg_index);
      free(append_data);
    }
  else
    {
      /* When append_data could not be allocated, 
       * or if message could not be fetched,
       * append_message is the fetched message
       */
      message = append_message;
    }

  x_free(message->data);
  message->data = NULL;
  message->len = 0;

  if (data->count)
    return;

  A_LOGGING_TIME("Append messages Stop");

  user_cb   = data->user_cb;
  user_data = data->user_data;
  failures  = data->failures;
  messages  = data->messages;
  folders   = data->folders;

  free(data);

  user_cb(folders, messages, failures, user_data);
}

void
a_Folder_append_file_async(XePFolder *folder, const char *filename,
			   XePManagerMsgCb user_cb, void *user_data)
{
  XePMessage *message;
  XePManagerMsg *msg;

  message = a_Message_new_from_file_noparse(filename);
  if (message)
    {
      /* Open the file so we can already safely unlink.
       * It will be closed in Folder_append_file_cb()
       */
      a_Message_file_open(message);
      a_Message_set_folder(message, folder);

      msg = a_Manager_msg_new();
      if (msg)
	{
	  a_Manager_msg_set_type(msg, XEP_MANAGER_APPEND);
	  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
	  /* NOTE 1: if the arg_out of XEP_MANAGER_APPEND changes
	   * thne change the arguments of user_cb below at NOTE 2.
	   */
	  a_Manager_msg_set_args(msg, message, message);
	  a_Manager_msg_add_callback(msg, Folder_append_file_cb, NULL);
	  if (user_cb)
	    {
	      a_Manager_msg_add_callback(msg, user_cb, user_data);
	    }
	  a_Manager_msg_submit(msg);
	}
      else
	{
	  A_LOGGING_ERROR("Message not sent.");
	  /* NOTE 2: see NOTE 1 */
	  if (user_cb)
	    user_cb(user_data, message);
	}
    }
}

static void
Folder_append_file_cb(void *anull, void *amessage)
{
  /* This will also close the opened tempfile */
  a_Message_unref((XePMessage *) amessage);
}

void
a_Folder_disconnect_async(XePFolder *folder, 
			  XePManagerMsgCb user_cb, void *user_data)
{
  
  XePManagerMsg *msg;

  if (a_Folder_get_thread_state(folder) == XEP_MANAGER_THREAD_STATE_NONE)
    {
      if (user_cb)
	{
	  user_cb(user_data, folder);
	}
    }
  else
    {
      msg = a_Manager_msg_new();
      a_Manager_msg_set_type(msg, XEP_MANAGER_DISCONNECT);
      a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
      a_Manager_msg_set_args(msg, folder, folder);
      if (user_cb)
	{
	  a_Manager_msg_add_callback(msg, user_cb, user_data);
	}
      a_Manager_msg_submit(msg);
    }
}

uint32_t
a_Folder_tree_get_total_count(XePFolder *folder)
{

  return (folder->ep_tree ? Folder_tree_get_total_count(folder->ep_tree) : 0);

}

static uint32_t
Folder_tree_get_total_count(struct mailmessage_tree *msg_tree)
{
  uint32_t n = 0;
  uint32_t i = 0;
  uint32_t t = 0;
  t = n = carray_count(msg_tree->node_children);
  for (i = 0; i < n; i++)
    {
      t += Folder_tree_get_total_count(carray_get(msg_tree->node_children, i));
    }
  return t;
}

void
a_Folder_messages_change_flags_async(XePFolder *folder, 
				     carray *messages,
				     XePManagerMsgCb user_cb,
				     void *user_data)
{
  XePManagerMsg *msg;

  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_LIST_FLAGS_CHANGE);
  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
  a_Manager_msg_set_args(msg, messages, messages);
  if (user_cb)
    a_Manager_msg_add_callback(msg, user_cb, user_data);
  a_Manager_msg_submit(msg);
}

XePFolderProcessing *
a_Folder_processing_new(XePFolder *folder, 
			carray *messages,
			clist *processing_rules)
{
  XePFolderProcessing *folder_processing;

  folder_processing = x_alloc(1, XePFolderProcessing);
  if (folder_processing == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }
  
  folder_processing->folder = folder;
  folder_processing->messages = messages;
  folder_processing->processing_rules = processing_rules;
  folder_processing->matches = NULL;
  return folder_processing;
}

void
a_Folder_processing_free(XePFolderProcessing *folder_processing)
{
  a_Message_array_free(folder_processing->messages);
  if (folder_processing->matches)
    clist_free(folder_processing->matches);
  free(folder_processing);
}

void
a_Folder_processing(XePFolderProcessing *folder_processing)
{
  carray *messages = folder_processing->messages;
  XePMessage *message;
  clist *processing_rules = folder_processing->processing_rules;
  XePProcessingRule *rule;
  clist *matches;
  uint32_t i;
  uint32_t count = carray_count(messages);
  clistiter *iter;
  int r;

#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  XePManagerThread *thread = &folder_processing->folder->thread;
  /* Get GIL (Global Interpreter Lock) */

  thread->py_thread_state = a_XePPython_thread_state_new();
#endif

  matches = clist_new();

  A_LOGGING_TIME("Processing Start");
  for (i = 0; i < count; i++)
    {
      message = (XePMessage *) carray_get(messages, i);
      
      for (iter = clist_begin(processing_rules); iter; iter = iter->next)
	{
	  rule = (XePProcessingRule *) iter->data;
	  r = a_Processing_message_match(rule->condition->tree, message);
	  if (r)
	    {
	      a_Processing_message_action(rule->action, message);
	      clist_append(matches, message);
	    }
	}
    }
  A_LOGGING_TIME("Processing Stop");
  printf("Matches: %d\n", clist_count(matches));
  if (clist_count(matches) == 0)
    {
      clist_free(matches);
      matches = NULL;
    }
  folder_processing->matches = matches;

  
#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  PyThreadState_Clear(thread->py_thread_state);
  PyThreadState_Delete(thread->py_thread_state);  
  thread->py_thread_state = NULL;
#endif
}

clist *
a_Folder_get_processing_rules(XePFolder *folder)
{
  clist *rules;
  XePProcessingRule *rule;
  XePProcessingRef *rule_ref;
  clistiter *iter;

  rules = clist_new();
  for (iter = clist_begin(folder->rule_refs); iter; iter = iter->next)
    {
      rule_ref = (XePProcessingRef *) iter->data;
      if (rule_ref->enabled)
	{
	  rule = a_Processing_rule_get_from_ref(rule_ref->ref);
	  if (rule)
	    clist_append(rules, rule);
	}
    }
  if (clist_count(rules) == 0)
    {
      clist_free(rules);
      rules = NULL;
    }
  return rules;  
}

void
a_Folder_processing_rule_add(XePFolder *folder, const char *ref)
{
  XePProcessingRef *rule_ref;
  clistiter *iter;

  for (iter = clist_begin(folder->rule_refs); iter; iter = iter->next)
    {
      rule_ref = (XePProcessingRef *) iter->data;
      if (strcmp(rule_ref->ref, ref) == 0)
	{
	  A_LOGGING_ERROR(_("Rule <%s> already present in folder <%s>"),
			  ref, folder->id);
	  return;
	}
    }
     
  rule_ref = a_Processing_ref_new();
  rule_ref->ref = strdup(ref);

  clist_append(folder->rule_refs, rule_ref);
}

void
a_Folder_processing_rule_remove(XePFolder *folder, const char *ref)
{
 XePProcessingRef *rule_ref;
  clistiter *iter;

  for (iter = clist_begin(folder->rule_refs); iter; iter = iter->next)
    {
      rule_ref = (XePProcessingRef *) iter->data;
      if (strcmp(rule_ref->ref, ref) == 0)
	{
	  a_Processing_ref_free(rule_ref);
	  clist_delete(folder->rule_refs, iter);
	  return;
	}
    }
  A_LOGGING_ERROR(_("No rule <%s> in folder <%s>"), ref, folder->id); 
}

/* Virtual folders */

void
a_Folder_virtual_process_messages(XePFolder *folder, carray *messages)
{
  unsigned int i;
  unsigned int count = carray_count(messages);
  struct mailmessage_tree *env_tree;
  int r;

  for (i = 0; i < count; i++)
    a_Folder_virtual_process_message(folder, carray_get(messages, i));
  
  /* FIXME: Do not hardcode the encoding (here UTF8) */
  r = mail_build_thread(MAIL_THREAD_REFERENCES_NO_SUBJECT,
			"UTF-8", 
			folder->ep_list, 
			&env_tree,
			mailthread_tree_timecomp);
  if (r == MAIL_NO_ERROR)
    {
      if (folder->ep_tree)
	{
	  mailmessage_tree_free_recursive(folder->ep_tree);
	}
      folder->ep_tree = env_tree;
    }
  A_LOGGING_SYSTEM_ERROR("mail_build_thread(): %s", maildriver_strerror(r));
}

int
a_Folder_virtual_process_message(XePFolder *folder, XePMessage *message)
{
  const char *condition_ref;
  XePProcessingCondition *condition;
  clistiter *iter;
  int matched;
  XePVMessage *vmessage;

  /* This assumes folder has conditions */
  for (iter = clist_begin(folder->condition_refs); iter; iter = iter->next)
    {
      condition_ref = (const char *) iter->data;
      condition = a_Processing_condition_get_from_ref(condition_ref);
      if (a_Processing_message_match(condition->tree, message))
	break;	  
    }

  matched = (iter != NULL);
  
  vmessage = a_VMessage_get_from_uid(message->folder, message->uid);

  if (matched)
    {
      if (!vmessage)
	{
	  vmessage = a_VMessage_new_from_message(message);
	  a_VMessage_ref_folder(vmessage, folder);
	  carray_add(folder->ep_list, message->msg, NULL);
	}
      return 0;
    }
  else
    {
      if (vmessage)
	{
	  /* XXX: efficiently remove a vmessage from an array:
	   * to remove from an array: put the vmessage with greatest
	   * index in place of the removed one.
	   * To avoid searching for the index of the to-be-removed
	   * message, we store it in the attribute structure 
	   */
	     
	  
	  a_VMessage_unref_folder(vmessage, folder);
	}
      return -1;
    }
}
