/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "lxml.h"

#include "intl.h"

#include "logging.h"
#include "x_alloc.h"
#include "foldersview.h"
#include "folder.h"
#include "mailbox.h"
#include "identity.h"

#define XEP_FOLDERSVIEW_MAILBOXES_ID _("All")

static void
Foldersview_tree_free_cb(cntree *atree, void *data);

static int
Foldersview_foldersview_cb(cntree *atree, void **data, void **data_iter, 
			   int op);

XePFoldersViewNode *
Foldersview_node_new_full(XePFolder *folder, const char *name);

static int
Foldersview_node_cb(cntree *atree, void **data, void **data_iter, int op);

static void *
Foldersview_list_begin_iter(XePFoldersViewsList *foldersviewslist);

static void *
Foldersview_list_next_iter(XePFoldersViewsList *foldersviewslist, void *iter);

static XePFoldersView *
Foldersview_list_get_from_iter(XePFoldersViewsList *foldersviewslist, 
			       void *iter);

static void
Foldersview_list_expunge_tree(cntree *tree);

static int
Foldersview_tree_link_folders(cntree *atree, void *data);

/* Static global variable */
/* TODO: remove all reference to the non-global foldersviewslist */

static XePFoldersViewsList *global_foldersviewslist;

/* Functions definitions */

XePFoldersView *
a_Foldersview_new(void)
{
  XePFoldersView * foldersview;
 
  foldersview = x_alloc(1, XePFoldersView);

  if (foldersview == NULL)
    A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
  else
    {
      foldersview->id = NULL;
      foldersview->tree = NULL;
      foldersview->lastviewed = NULL;
    }
  return foldersview;
}

void
a_Foldersview_free(XePFoldersView *foldersview)
{
  if (foldersview == NULL)
    return;

  cntree_free(foldersview->tree, Foldersview_tree_free_cb, NULL);

  x_free(foldersview->id);
  x_free(foldersview);
}

static void
Foldersview_tree_free_cb(cntree *atree, void *data)
{
  if (atree && atree->data)
    a_Foldersview_node_free((XePFoldersViewNode *) atree->data);
}

XePFoldersViewNode *
a_Foldersview_node_new(void)
{
  return Foldersview_node_new_full(NULL, NULL);
}

XePFoldersViewNode *
Foldersview_node_new_full(XePFolder *folder, const char *id)
{
  XePFoldersViewNode *node;

  node = x_alloc(1, XePFoldersViewNode);
  if (node)
    {
      node->folder_ref.folder = folder;
      node->folder_ref.ref = folder ? X_STRDUP(a_Folder_get_id(folder)) : NULL;
      node->name = X_STRDUP(id);
      node->property = XETPAN_FOLDERSVIEW_NODE_VISIBLE;
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
    }

  return node;
}

void
a_Foldersview_node_free(XePFoldersViewNode *node)
{
  if (node == NULL)
    return;

  x_free(node->folder_ref.ref);
  x_free(node->name);

  x_free(node);
}

static int
Foldersview_mailboxes_view_new(XePFoldersViewsList *foldersviewslist)
{
  int r = -1;

  foldersviewslist->mailboxes = a_Foldersview_new();
  if (foldersviewslist->mailboxes)
    {
      /* FIXME: mailboxes->id seems to be not freeed */
      foldersviewslist->mailboxes->id = strdup(XEP_FOLDERSVIEW_MAILBOXES_ID);
      if (foldersviewslist->mailboxes->id)
	{
	  r = 0;
	}
      else
	{
	  A_LOGGING_SYSTEM_ERROR("strdup(): %s", strerror(errno));
	  a_Foldersview_free(foldersviewslist->mailboxes);
	  foldersviewslist->mailboxes = NULL;
	}
    }
  return r;
}

int
a_Foldersview_mailboxes_view_update(void)
{
  clistiter *id_iter;
  clistiter *mb_iter;
  clistiter *iter;
  XePIdentity *identity;
  cntree *parent;
  cntree *child;
  cntree *gd_child;
  XePMailbox *mailbox;
  XePFolder *folder;
  XePFoldersViewNode *node;
  clist *identities;
  const char *name;

  /* Delete the default view of all mailboxes ("All") */
  a_Foldersview_free(global_foldersviewslist->mailboxes);

  if (Foldersview_mailboxes_view_new(global_foldersviewslist) != 0)
    {
      return -1;
    }

  identities = a_Identity_list_get();

  parent = cntree_new();
  if (parent == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return -1;
    }

  global_foldersviewslist->mailboxes->tree = parent;

  for (id_iter = clist_begin(identities); id_iter; id_iter = id_iter->next)
    {
      identity = (XePIdentity *) id_iter->data;
      for (mb_iter = clist_begin(identity->mailboxes); 
	   mb_iter; 
	   mb_iter = mb_iter->next)
	{
	  mailbox = (XePMailbox *) mb_iter->data;
	  
	  node = Foldersview_node_new_full(mailbox->folder, 
					   a_Folder_get_id(mailbox->folder));
	  child = cntree_new();
	  child->data = node;
	  cntree_append_child(parent, child);
	  for (iter = clist_begin(mailbox->folders); iter; iter = iter->next)
	    {
	      /* TODO: reflect IMAP or MH hiearchies */
	      folder = (XePFolder *) iter->data;
	      name = a_Folder_get_short_name(folder);
	      node = Foldersview_node_new_full(folder, name);
	      gd_child = cntree_new();
	      gd_child->data = node;
	      cntree_append_child(child, gd_child);
	    }
	}
    }
  return 0;
}

int
a_Foldersview_list_create(cntree *ntree,
		     XePFoldersViewsList **foldersviewslist,
		     int op)
{
  int r;
  LXMLelement *element;
  LXMLelement_callbacks element_cb[] = 
    {
      {"foldersview", Foldersview_foldersview_cb, NULL},
      {NULL, NULL, NULL}
    };
  
  element = lxml_tree_get_element(ntree);

  element_cb[0].data = (void **) foldersviewslist;

  A_LOGGING_INFO("Found foldersviewslist. Parsing...");
  r = lxml_tree_convert(ntree, element_cb, op);
  A_LOGGING_INFO("Finished foldersviewslist parsing.");

  return r;
}

static int
Foldersview_foldersview_cb(cntree *atree, void **data, void **data_iter,
			   int op)
{
  int r;
  LXMLelement *element;
  XePFoldersViewsList *foldersviewslist = *((XePFoldersViewsList **) data);
  XePFoldersView *foldersview;
  LXMLstore store[] = 
    {
      {"id", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  LXMLelement_callbacks element_cb[] = 
    {
      {"folder_ref", Foldersview_node_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if ((NULL == *data_iter) &&
	  (NULL == (*data_iter = Foldersview_list_begin_iter(foldersviewslist))))
	return 0;
      
      foldersview = Foldersview_list_get_from_iter(foldersviewslist, 
						   *data_iter);
      *data_iter = Foldersview_list_next_iter(foldersviewslist, *data_iter);
    }
  else
    {
      foldersview = a_Foldersview_new();
      if (foldersview == NULL)
	return -1;
    }
  
  store[0].dest = (void **)&(foldersview->id);

  r= lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Foldersview syntax error. Ignored.");
	  goto err;
	}
      
      if (!foldersview->id)
	{
	  A_LOGGING_ERROR("Foldersview does not have an ID. Ignored.");
	  goto err;
	}
      
      foldersview->tree = cntree_new();
      if (foldersview->tree == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("cntree_new failed");
	  goto err;
	}
    }

  element_cb[0].data = (void **) &(foldersview->tree);
  
  A_LOGGING_INFO("Found foldersview. Parsing...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_INFO("Finished foldersview parsing");

  if (op == LXML_Read && r == 0)
    {
      clist_append(foldersviewslist->list, foldersview);
    }

  return r;
  
 err:
  a_Foldersview_free(foldersview);
  return -1;
}

static int
Foldersview_node_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  cntree *parent = *((cntree **) data);
  cntree *achild;
  XePFoldersViewNode *node;
  LXMLelement *element;
  LXMLelement_callbacks element_cb[] =
    {
      {"folder_ref", Foldersview_node_cb, NULL},
      {NULL, NULL, NULL}
    };
  LXMLstore store[] = 
    {
      {"ref",     LXML_String,  NULL, LXML_Unique},
      {"name",    LXML_String,  NULL, LXML_Unique},
      {"enabled", LXML_Boolean, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if (NULL == *data_iter)
	{
	  if (parent->children == NULL)
	    return 0;
	  *data_iter = clist_begin(parent->children);
	  if (NULL == *data_iter)
	    return 0;
	}
      achild = (cntree *) (((clistiter *) *data_iter)->data);
      *data_iter = ((clistiter *) *data_iter)->next;
      node = (XePFoldersViewNode *) achild->data;
    }
  else
    {
      node = a_Foldersview_node_new();
      if (node == NULL)
	goto err;
    }

  store[0].dest = (void **)&node->folder_ref.ref;
  store[1].dest = (void **)&node->name;
  store[2].dest = (void **)&node->enabled;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Foldersview's folder syntax error. Ignored.");
	  goto err;
	}
      
      achild = cntree_new();
      if (achild == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("cntree_new failed");
	  goto err;
	}
      achild->data = node;
    }

  element_cb[0].data = (void **) &achild;

  r = lxml_tree_convert(atree, element_cb, op);
  
  if (op == LXML_Read)
    {
      if (r == 0)
	cntree_append_child(parent, achild);
      else
	{
	  cntree_free(achild, NULL, NULL);
	  goto err;
	}
    }

  return r;

 err:
  a_Foldersview_node_free(node);
  return -1;
}
 
XePFoldersViewsList *
a_Foldersview_list_new(void)
{
  XePFoldersViewsList *foldersviewslist;

  foldersviewslist = x_alloc(1, XePFoldersViewsList);

  if (foldersviewslist)
    {
      foldersviewslist->list = clist_new();
      if (foldersviewslist->list)
	{
	  if (Foldersview_mailboxes_view_new(foldersviewslist) != 0)
	    {
	      clist_free(foldersviewslist->list);
	      free(foldersviewslist);
	      foldersviewslist = NULL;
	    }
	}
      else
	{
	  A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
	  free(foldersviewslist);
	  foldersviewslist = NULL;
	}
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
    }

  global_foldersviewslist = foldersviewslist;
  return foldersviewslist;
}

void
a_Foldersview_list_free(XePFoldersViewsList *foldersviewslist)
{
  clistiter *iter;
  XePFoldersView *foldersview;

  if (foldersviewslist == NULL)
    return;

  for (iter = clist_begin(foldersviewslist->list); iter; iter = iter->next)
    {
      foldersview = (XePFoldersView *) iter->data;
      a_Foldersview_free(foldersview);
    }

  clist_free(foldersviewslist->list);

  a_Foldersview_free(foldersviewslist->mailboxes);

  free(foldersviewslist);

  global_foldersviewslist = NULL;

  return;
}

static void *
Foldersview_list_begin_iter(XePFoldersViewsList *foldersviewslist)
{
  return clist_begin(foldersviewslist->list);
}

static void *
Foldersview_list_next_iter(XePFoldersViewsList *foldersviewslist, void *iter)
{
  return ((clistiter *) iter)->next;
}

static XePFoldersView *
Foldersview_list_get_from_iter(XePFoldersViewsList *foldersviewslist, 
			       void *iter)
{
  return (XePFoldersView *) ((clistiter *) iter)->data;
}

void
a_Foldersview_list_expunge(void)
{
  clistiter *iter;
  XePFoldersView *foldersview;
  int ret = 0;
  
  for (iter = clist_begin(global_foldersviewslist->list);
       iter && ret == 0;
       iter = iter->next)
    {
      foldersview = (XePFoldersView *) iter->data;
      if (foldersview->tree)
	{
	  Foldersview_list_expunge_tree(foldersview->tree);
	}
    }
}

static void
Foldersview_list_expunge_tree(cntree *tree)
{
  cntree *child;
  clistiter *iter;
  XePFoldersViewNode *node;
  int r;

  if (tree == NULL || tree->children == NULL)
    return;

  iter = clist_begin(tree->children);
  while (iter)
    {
      clistiter *next = iter->next;

      child = (cntree *) iter->data;

      /* Expunge child subtree */
      Foldersview_list_expunge_tree(child);

      node = (XePFoldersViewNode *) child->data;

      if (node->folder_ref.ref)
	{
	  r = a_Folder_ref_link_to_folder(&(node->folder_ref));
	  if (r)
	    {
	      /* Folder ref is referring to a no more existing folder */
	      node->folder_ref.folder = NULL;
	      /* Remove it only if it has no valid children */
	      if (cntree_get_children_count(child) == 0)
		{
		  cntree_remove_child(tree, child);
		  a_Foldersview_node_free(node);
		  cntree_free(child, NULL, NULL);
		}
	    }
	}
      
      iter = next;
    }
}

int
a_Foldersview_list_link_folders(XePFoldersViewsList *foldersviewslist)
{
  clistiter *iter;
  XePFoldersView *foldersview;
  int ret = 0;

  A_LOGGING_TRACE;

  for (iter = clist_begin(foldersviewslist->list); 
       iter && !ret; 
       iter = iter->next)
    {
      foldersview = (XePFoldersView *) iter->data;
      if (foldersview->tree)
	ret = cntree_for_each(foldersview->tree, 
			      Foldersview_tree_link_folders,
			      NULL);
      if (ret)
	A_LOGGING_ERROR("foldersview %s could not be linked to its folders",
			foldersview->id);
    }

  /* TODO: link mailboxes */

  return ret;
}

static int
Foldersview_tree_link_folders(cntree *atree, void *data)
{
  XePFoldersViewNode *foldersviewnode;
  int ret = 0;

  A_LOGGING_TRACE;

  if (!atree->data)
      return 0;

  foldersviewnode = (XePFoldersViewNode *) atree->data;

  ret = a_Folder_ref_link_to_folder(&(foldersviewnode->folder_ref));
  if (ret)
    {
      A_LOGGING_ERROR("Reference made to a non existent folder '%s'",
 		      foldersviewnode->folder_ref.ref);
    }

  return ret;
}


XePFoldersView *
a_Foldersview_list_get_foldersview(XePFoldersViewsList *foldersviewslist,
				   const char *foldersview_id)
{
  clistiter *iter;
  XePFoldersView *foldersview;

  if (foldersview_id && 
      strcmp(foldersview_id, XEP_FOLDERSVIEW_MAILBOXES_ID) == 0)
    return foldersviewslist->mailboxes;

  for (iter = clist_begin(foldersviewslist->list); iter; iter = iter->next)
    {
      foldersview = (XePFoldersView *) iter->data;
      if (!(foldersview_id  && *foldersview_id) ||
	  strcmp(foldersview->id, foldersview_id) == 0)
	return foldersview;
    }

  return foldersviewslist->mailboxes;
}

static int
Foldersview_node_dump(cntree *atree, void *anindent)
{
  int indent = *((int *) anindent);
  int i;
  XePFoldersViewNode *node = (XePFoldersViewNode *) atree->data;

  indent += 2;
  for (i = 0; i < indent; i++)
    printf(" ");

  if (node)
    {
      cntree_for_each_child(atree, Foldersview_node_dump, &indent);
    }
  return 0;
}


static void
Foldersview_dump(void *afoldersview, void *data)
{
  static int i = 0;
  int indent = 0;
  XePFoldersView *foldersview = (XePFoldersView *) afoldersview;

  printf("FoldersView %d id = %s\n", i++, foldersview->id);
  cntree_for_each_child(foldersview->tree, Foldersview_node_dump, &indent);
}

static void
Foldersview_list_dump(XePFoldersViewsList *foldersviewslist)
{
  clist_foreach((clist *) foldersviewslist->list, Foldersview_dump, NULL);
  Foldersview_dump(foldersviewslist->mailboxes, NULL);
}
