/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI
#ifdef USE_PYTHON

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <gtk/gtkwidget.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtktextbuffer.h>
#include <gtk/gtktextview.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkvpaned.h>
#include <gdk/gdkkeysyms.h>
#include <Python.h>

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"

#include "ui.h"
#include "xep-python.h"
#include "gtk-utils.h"

#include "gtk-python.h"

#define PYTHON_CONSOLE_WELCOME "*** This is XetPan's Python Console ***\n"


struct XePGTKPython_s
{
  GtkWidget *window;
  GtkTextView *textview;
  GtkTextBuffer *buffer;
  GtkTextView *i_textview;
  GtkTextBuffer *i_buffer;
#if 0
  GtkTextTag *noteditable_tag;
#endif
  GtkTextTag *pystdout_tag;
  GtkTextTag *pystderr_tag;
  GtkTextMark *text_end_mark;
  int pystdin[2];
  int pystdout[2];
  int pystderr[2];
  unsigned int pystdout_wid;
  unsigned int pystderr_wid;
  GQueue *history;
  GList *history_item;
  XetPan *xetpan;
};

/* Global variable: there must be one and only one Python interpreter
 * (console) running at the same time
 */
static XePGTKPython *global_xeppy;



/* Python Console */

static int
GtkPython_initialize(XePGTKPython *xeppy);

static void
GtkPython_init_module(void);

static void
GtkPython_finish(XePGTKPython *xeppy);

static void
GtkPython_free(XePGTKPython *xeppy);

static void
GtkPython_restart_session_clicked_cb(GtkButton *button, gpointer axeppy);

static gboolean
GtkPython_delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data);

static void
GtkPython_destroy_cb(GtkWidget *widget, gpointer data);

static gboolean
GtkPython_key_press_cb(GtkWidget *widget, GdkEventKey *event, gpointer data);

static PyObject *
GtkPython_write_stdout(PyObject *self, PyObject *args);

static PyObject *
GtkPython_write_stderr(PyObject *self, PyObject *args);

static void
GtkPython_text_buffer_append_output(XePGTKPython *xeppy, 
				    const gchar *text,
				    gboolean is_stderr);

static void
GtkPython_prepare_console(XePGTKPython *xeppy);

static PyMethodDef xep_gtkpython_module_def[] =
  {
    {
      "xep_gtkpython_write_stdout", 
      GtkPython_write_stdout, 
      METH_VARARGS, 
      "Write normal output to XetPan's Gtk+/Python console"
    },
    {
      "xep_gtkpython_write_stderr", 
      GtkPython_write_stderr, 
      METH_VARARGS, 
      "Write error messages to XetPan's Gtk+/Python console"
    },
    {NULL, NULL, 0, NULL}
  };


/* Function definitions */

XePGTKPython *
a_GtkPython_console_new(XetPan *xetpan, GtkWidget *toplevel)
{
  GtkWidget *window;
  GtkWidget *box;
  GtkWidget *i_box;
  GtkWidget *scrollw;
  GtkWidget *textview;
  GtkWidget *i_textview;
  GtkWidget *button;
  GtkWidget *pane;
  GtkWidget *label;
  XePGTKPython *xeppy;
  GtkTextIter start, end;
  gchar *display;
  PangoFontDescription *fontdesc;

  if (a_XePPython_is_initialized())
    return NULL;

  xeppy = x_alloc(1, XePGTKPython);
  if (xeppy == NULL)
    return NULL;

  xeppy->xetpan = xetpan;
  xeppy->history = g_queue_new();

  display =  gdk_screen_make_display_name(gtk_window_get_screen
					  (GTK_WINDOW(toplevel)));

  window = a_GtkUtils_toplevel_new(display,
				 _("XetPan - Python Console"),
				   NULL,
				   500, 300);
  box = gtk_vbox_new(FALSE, 6);
  gtk_container_set_border_width(GTK_CONTAINER(box), 12);
  gtk_container_add(GTK_CONTAINER(window), box);

  label = a_GtkUtils_label_new_bold(_("Python Console"));
  gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, 0);

  pane = gtk_vpaned_new();
  gtk_box_pack_start(GTK_BOX(box), pane, TRUE, TRUE, 0);

  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_container_set_border_width(GTK_CONTAINER(scrollw), 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw),
				      GTK_SHADOW_IN);
  gtk_paned_pack1(GTK_PANED(pane), scrollw, TRUE, FALSE);

  textview = a_GtkUtils_text_view_new(NULL, -1, FALSE);
  /* Show cursor to allow keyboard navigation and selection */
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(textview), TRUE);

  gtk_container_add(GTK_CONTAINER(scrollw), textview);

  fontdesc = pango_font_description_from_string("monospace");
  gtk_widget_modify_font(textview, fontdesc);

  i_box = gtk_vbox_new(FALSE, 0);
  label = a_GtkUtils_label_new_bold(_("Input Console"));
  gtk_box_pack_start(GTK_BOX(i_box), label, FALSE, FALSE, 6);
  gtk_paned_pack2(GTK_PANED(pane), i_box, FALSE, FALSE);

  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_container_set_border_width(GTK_CONTAINER(scrollw), 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw), 
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollw),
				      GTK_SHADOW_IN);
  gtk_box_pack_start(GTK_BOX(i_box), scrollw, TRUE, TRUE, 0);

  i_textview = a_GtkUtils_text_view_new(NULL, -1, TRUE);
  gtk_container_add(GTK_CONTAINER(scrollw), i_textview);

  gtk_widget_modify_font(i_textview, fontdesc);
  pango_font_description_free(fontdesc);

  button = gtk_button_new_with_label("Restart Python Session");
  gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(button), "clicked", 
		   G_CALLBACK(GtkPython_restart_session_clicked_cb), 
		   xeppy);

  xeppy->window = window;
  xeppy->textview = GTK_TEXT_VIEW(textview);
  xeppy->buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(textview));
  xeppy->i_textview = GTK_TEXT_VIEW(i_textview);
  xeppy->i_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(i_textview));

#if 0
  xeppy->noteditable_tag = gtk_text_buffer_create_tag(xeppy->buffer, 
						      "not-editable", 
						      "editable", FALSE,
						      NULL);
#endif
  xeppy->pystdout_tag = gtk_text_buffer_create_tag(xeppy->buffer, 
						   "pystdout", 
						   "foreground", "blue",
						   NULL);
  xeppy->pystderr_tag = gtk_text_buffer_create_tag(xeppy->buffer, 
						   "pystderr", 
						   "foreground", "red",
						   NULL);

  gtk_text_buffer_get_bounds(xeppy->buffer, &start, &end);
#if 0
  gtk_text_buffer_apply_tag(xeppy->buffer, xeppy->noteditable_tag,
			    &start, &end);
#endif
  xeppy->text_end_mark = gtk_text_buffer_create_mark(xeppy->buffer, NULL,
						     &end, FALSE);

  /*
    gtk_text_view_scroll_to_mark(xeppy->textview, xeppy->text_end_mark, 0.0,
    FALSE, 0.0, 0.0);
    gtk_text_buffer_place_cursor(xeppy->buffer, &end);
  */

  /* Signals */
  
  g_signal_connect(G_OBJECT(window), "destroy",
		   G_CALLBACK(GtkPython_destroy_cb), xeppy);

  g_signal_connect(G_OBJECT(window), "delete-event",
		   G_CALLBACK(GtkPython_delete_cb), xeppy);

  g_signal_connect(G_OBJECT(i_textview), "key-press-event",
		   G_CALLBACK(GtkPython_key_press_cb), xeppy);
#if 0
  g_signal_connect(G_OBJECT(button), "clicked",
		   G_CALLBACK(GtkPython_delete_cb), xeppy);
#endif

  /* Focus */
  gtk_widget_grab_focus(i_textview);
  gtk_widget_show_all(window);

  xeppy->pystdin[0]  = 0;
  xeppy->pystdin[1]  = 0;
  xeppy->pystdout[0] = 0;
  xeppy->pystdout[1] = 0;
  xeppy->pystderr[0] = 0;
  xeppy->pystderr[1] = 0;
  xeppy->pystdout_wid = 0;
  xeppy->pystderr_wid = 0;

  global_xeppy = xeppy;

  if (GtkPython_initialize(xeppy))
    {
      GtkPython_finish(xeppy);
      xeppy = NULL;
    }
  GtkPython_text_buffer_append_output(xeppy, PYTHON_CONSOLE_WELCOME, FALSE);
  GtkPython_prepare_console(xeppy);
  return xeppy;
}

static int
GtkPython_initialize(XePGTKPython *xeppy)
{

  if (pipe(xeppy->pystdout))
    {
      A_LOGGING_SYSTEM_ERROR("pipe: %s\nPython not initialized.",
			     strerror(errno));
      return -1;
    }

  a_XePPython_initialize();

  GtkPython_init_module();

  return 0;     
}

static void
GtkPython_finish(XePGTKPython *xeppy)
{
  gtk_widget_destroy(xeppy->window);
}

static void
GtkPython_free(XePGTKPython *xeppy)
{
  GList *iter;
  char *text;

  a_XePPython_finalize();

#if 0
  a_UI_remove_io_watch(xeppy->xetpan, xeppy->pystdout_wid);
  a_UI_remove_io_watch(xeppy->xetpan, xeppy->pystderr_wid);
#endif

  /* Free history */
  for (iter = xeppy->history->head; iter; iter = iter->next)
    {
      text = (char *) iter->data;
      x_free(text);
    }
  g_queue_free(xeppy->history);

  close(xeppy->pystdout[0]);
  close(xeppy->pystdout[1]);
  close(xeppy->pystderr[0]);
  close(xeppy->pystderr[1]);

  
  x_free(xeppy);

  global_xeppy = NULL;

}

static void
GtkPython_init_module(void)
{
  PyObject *gtkpython_module;
  char *gtkpython_init_code =
    "class XePGtkPythonStdout:\n"
    "    def write(self, s): \n"
    "        xep_gtkpython_write_stdout(s)\n"
    "\n"
    "class XePGtkPythonStderr:\n"
    "    def write(self, s): \n"
    "        xep_gtkpython_write_stderr(s)\n"
    "\n"
    "import sys\n"
    "from xep_gtkpython import *\n"
    "sys.stdout = XePGtkPythonStdout()\n"
    "sys.stderr = XePGtkPythonStderr()\n"
    "sys.stdin  = None\n";

#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  a_XePPython_thread_main_acquire();
#endif

  gtkpython_module = Py_InitModule("xep_gtkpython", xep_gtkpython_module_def);
  PyRun_SimpleString(gtkpython_init_code);

#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  a_XePPython_thread_main_release();
#endif
}

static PyObject *GtkPython_write_stdout(PyObject *self, PyObject *args)
{
  char *output;

  if (!PyArg_ParseTuple(args, "s", &output))
    {
      return NULL; /* Exception */
    }

  GtkPython_text_buffer_append_output(global_xeppy, output, FALSE);

  return Py_BuildValue("");
}

static PyObject *GtkPython_write_stderr(PyObject *self, PyObject *args)
{
  char *output;

  if (!PyArg_ParseTuple(args, "s", &output))
    {
      return NULL; /* Exception */
    }

  GtkPython_text_buffer_append_output(global_xeppy, output, TRUE);

  return Py_BuildValue("");
}

static void
GtkPython_text_buffer_append_output(XePGTKPython *xeppy, 
				    const gchar *text,
				    gboolean is_stderr)
{
  GtkTextTag *tag;
  GtkTextIter end_iter;

  if (is_stderr)
    {
      tag = xeppy->pystderr_tag;
    }
  else
    {
      tag = xeppy->pystdout_tag;
    }

  /* Insert text */
  gtk_text_buffer_get_end_iter(xeppy->buffer, &end_iter);
  gtk_text_buffer_insert_with_tags(xeppy->buffer, &end_iter,
				   text, -1, tag, NULL);

}

static void
GtkPython_restart_session_clicked_cb(GtkButton *button, gpointer axeppy)
{
  XePGTKPython *xeppy = (XePGTKPython *) axeppy;
  const char *msg = _("\n*** Python session restarted ***\n");
  GtkPython_text_buffer_append_output(xeppy, msg, TRUE);
  a_XePPython_finalize();
  GtkPython_initialize(xeppy);
  GtkPython_prepare_console(xeppy);
}

static gboolean
GtkPython_delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_widget_hide(widget);

  return TRUE;
}

static void
GtkPython_destroy_cb(GtkWidget *widget, gpointer data)
{
  XePGTKPython *xeppy = (XePGTKPython *) data;
  GtkPython_free(xeppy);
}

static gboolean
GtkPython_key_press_cb(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
  XePGTKPython *xeppy = (XePGTKPython *) data;
  if (((event->state & GDK_MODIFIER_MASK) == GDK_CONTROL_MASK) &&
      (event->keyval == GDK_Return || 
       event->keyval == GDK_Down ||
       event->keyval == GDK_Up))
    {
      GtkTextIter start_iter, end_iter;
      gchar *text;
      GList *next_item;
      

      gtk_text_buffer_get_bounds(xeppy->i_buffer, &start_iter, &end_iter);
      text = gtk_text_buffer_get_text(xeppy->i_buffer, 
				      &start_iter, &end_iter, FALSE);
      if (text && *text && 
	  (xeppy->history_item == xeppy->history->tail ||
	   event->keyval == GDK_Return))
	{
	  GList *tail;
	  tail = xeppy->history->tail;
	  x_free(tail->data);
	  tail->data = text;
	  /* Execute text */
	  if (event->keyval == GDK_Return)
	    {
	      /* Execute and insert Python output */
	      gtk_text_buffer_get_end_iter(xeppy->buffer, &end_iter);
	      gtk_text_buffer_insert(xeppy->buffer, &end_iter, text, -1);
	      gtk_text_buffer_insert(xeppy->buffer, &end_iter, "\n", -1);
	      a_XePPython_execute(text);
	      
	      GtkPython_prepare_console(xeppy);
	      return TRUE;
	    }
	}
      /* History navigation */
      next_item = (event->keyval == GDK_Down ? 
		   xeppy->history_item->next :
		   xeppy->history_item->prev);
      if (next_item)
	{
	  text = (char *) next_item->data;
	  gtk_text_buffer_set_text(xeppy->i_buffer, (text ? text : ""), -1);
	  xeppy->history_item = next_item;
	  return TRUE;
      	}
      /* If history did not succeed, then let Gtk2 process Ctrl-Up/Down */
    }
  return FALSE;
}

static void
GtkPython_prepare_console(XePGTKPython *xeppy)
{
  GtkTextIter end_iter;
  
  /* Make the whole buffer not-editable again */
  gtk_text_buffer_set_text(xeppy->i_buffer, "", -1);
  gtk_text_buffer_get_end_iter(xeppy->buffer, &end_iter);
  gtk_text_buffer_place_cursor(xeppy->buffer, &end_iter);
  gtk_text_view_scroll_to_mark(xeppy->textview, xeppy->text_end_mark, 0.0,
			       FALSE, 0.0, 0.0);

  if (xeppy->history->length > XEP_PYTHON_HISTORY_SIZE)
    {
      char *text;
      text = g_queue_pop_head(xeppy->history);
      x_free(text);
    }
  g_queue_push_tail(xeppy->history, NULL);
  xeppy->history_item = xeppy->history->tail;
}

void
a_GtkPython_console_show(XePGTKPython *xeppy)
{
  gtk_window_present(GTK_WINDOW(xeppy->window));
}

#endif /* USE_PYTHON */
#endif /* USE_GTKUI */
